/*
ManjusakaJS: Script generator
=============================
This script can be used either in CLI or in web page. It takes a YAML document
file and generates a "compiled" HTML code which stands alone for decryption.

The document file must be in following format:

questions:
    What is my name: John Doe
    What is my birthday:
        - hint: yyyy-mm-dd
        - answer: 1970-01-01
masterpassword: 8135782316578321091
---
Here goes the content...


The document file precedes with a YAML config file, ending with "---", and then
the main content. The config file must contain at least one of following entires:

1. questions
2. password

"questions" must be a dictionary, where each key is a question and as item
either plain string as answer, or another dictionary with "hint" and "answer"
is given.
"password" must be a string.

All question answers is used to derive a password to decrypt the main content.
If a master password is specified, both the questions(if present) and the
main content will be encrypted. This prevents leakage of questions prior to
any "authorized" access to this page.
*/
function isNodejs() {
    return (
        typeof "process" !== "undefined" &&
        process && process.versions && process.versions.node);
}

function isString(x){
    return Object.prototype.toString.call(x) === "[object String]"
}


const showdown = require("showdown");
const template = "{{ manjusaka-template }}";
const openpgp = require("./openpgp.min.js");
const yaml = require("js-yaml");
const buffer = require("buffer");
const generateQuestionsBasedKey = require("./questions-based-key.js");
const generatePasswordBasedKey = require("./password-based-key.js");



async function generator(doc, onProgress){
    const splitPos = doc.indexOf("\n----\n");
    if(splitPos < 0){
        throw Error("Invalid document.");
    }
    const docOptions = yaml.safeLoad(doc.slice(0, splitPos));

    const converter = new showdown.Converter();
    const docContent = converter.makeHtml(doc.slice(splitPos + 7));


    if(onProgress == undefined) onProgress = function(){};
    
    // To derive a questions-based key, at each page generation a random
    // nonce is added.
    const nonce = await openpgp.crypto.random.getRandomBytes(64);

    // read in options
    if(docOptions.password !== undefined && !isString(docOptions.password)){
        throw Error("Document option [password] must be a string.");
    }
    var usePassword = isString(docOptions.password);
    var strPassword = docOptions.password;
    var questions = {};
    var useQuestions = false;
    if(docOptions.questions !== undefined){
        useQuestions = true;
        for(var question in docOptions.questions){
            var definition = docOptions.questions[question];
            if(!isString(question)) throw Error("Question must be a string.");
            if(isString(definition)){
                questions[question] = {
                    answer: definition,
                    hint: false,
                };
            } else {
                if( 
                    !isString(definition["answer"]) ||
                    (
                        definition["hint"] !== undefined &&
                        !isString(definition["hint"])
                    )
                ){
                    throw Error("Question definition is invalid.");
                }
                questions[question] = {
                    answer: definition["answer"],
                    hint: definition["hint"] || false,
                }
            }
        }
    }

    // encryptionless is disabled
    if(!useQuestions && !usePassword){
        throw Error("Either questions or master password must be set.");
    }

    // now generate outputs
    
    const questionsBasedKey = (useQuestions ? (
        await generateQuestionsBasedKey(nonce, questions, onProgress)):"");

    const passwordBasedKey = (usePassword ? (
        await generatePasswordBasedKey(nonce, strPassword, onProgress)):"");

    const contentEncryptionKey = questionsBasedKey + "|" + passwordBasedKey;
    const encryptedContent = (await openpgp.encrypt({
        passwords: [contentEncryptionKey],
        message: openpgp.message.fromText(docContent),
    })).data;

    var questionHints = {};
    for(var name in questions){
        questionHints[name] = questions[name].hint || false;
    };
    if(usePassword && useQuestions){
        questionHints = (await openpgp.encrypt({
            passwords: [passwordBasedKey],
            message: openpgp.message.fromText(JSON.stringify(questionHints)),
        })).data;
    }

    const replaceTable = {
        "{{ manjusaka-js-payload-N }}": JSON.stringify(generateQuestionsBasedKey.N),
        "{{ manjusaka-js-payload-nonce }}": JSON.stringify(new buffer.Buffer(nonce).toString("base64")),
        "{{ manjusaka-js-payload-hints }}": JSON.stringify(questionHints),
        "{{ manjusaka-js-payload-content }}": JSON.stringify(encryptedContent),
    };
    var output = template;
    for(var f in replaceTable){
        var v = replaceTable[f];
        output = output.replace("\"" + f + "\"", v);
    }

    return output;
}


module.exports = generator;
